@@ -48,22 +48,22 @@
             [com.puppetlabs.puppetdb.command.dlo :as dlo]
             [com.puppetlabs.puppetdb.query.population :as pop]
             [com.puppetlabs.jdbc :as pl-jdbc]
-            [com.puppetlabs.jetty :as jetty]
             [com.puppetlabs.mq :as mq]
-            [puppetlabs.kitchensink.core :as kitchensink]
             [clojure.java.jdbc :as sql]
             [clojure.string :as string]
             [clojure.tools.logging :as log]
             [com.puppetlabs.cheshire :as json]
             [com.puppetlabs.puppetdb.http.server :as server]
             [com.puppetlabs.puppetdb.config :as conf]
-            [com.puppetlabs.puppetdb.utils :as utils])
+            [com.puppetlabs.puppetdb.utils :as utils]
+            [puppetlabs.kitchensink.core :as kitchensink]
+            [puppetlabs.trapperkeeper.core :refer [defservice main]]
+            [compojure.core :as compojure])
   (:use [clojure.java.io :only [file]]
         [clj-time.core :only [ago secs minutes days]]
         [overtone.at-at :only (mk-pool interspaced)]
         [com.puppetlabs.time :only [to-secs to-millis parse-period format-period period?]]
         [com.puppetlabs.jdbc :only (with-transacted-connection)]
-        [puppetlabs.kitchensink.core :only (cli! with-error-delivery)]
         [com.puppetlabs.repl :only (start-repl)]
         [com.puppetlabs.puppetdb.scf.migrate :only [migrate!]]
         [com.puppetlabs.puppetdb.version :only [version update-info]]
@@ -76,7 +76,6 @@
 ;; The following functions setup interaction between the main
 ;; PuppetDB components.
 
-(def configuration nil)
 (def mq-addr "vm://localhost?jms.prefetchPolicy.all=1&create=false")
 (def mq-endpoint "com.puppetlabs.puppetdb.commands")
 (def send-command! (partial command/enqueue-command! mq-addr mq-endpoint))
@@ -94,7 +93,7 @@
      (log/error exception "Error during command processing; reestablishing connection after 10s")
      (Thread/sleep 10000))
 
-   (with-open [conn (mq/connect! mq)]
+   (with-open [conn (mq/activemq-connection mq)]
      (command/process-commands! conn mq-endpoint discard-dir opt-map))))
 
 (defn auto-deactivate-nodes!
@@ -214,31 +213,54 @@
           (log/warnf "%s rejected by certificate whitelist %s" ssl-client-cn whitelist)
           false)))))
 
-(defn on-shutdown
-  "General cleanup when a shutdown request is received."
-  []
-  ;; nothing much to do here for now, but let's at least log that we're shutting down.
-  (log/info "Shutdown request received; puppetdb exiting."))
+(defn shutdown-mq-broker
+  "Explicitly shut down the queue `broker`"
+  [{:keys [broker]}]
+  (when broker
+    (log/info "Shutting down message broker.")
+    ;; Stop the mq the old-fashioned way
+    (mq/stop-broker! broker)))
 
-(def supported-cli-options
-  [["-c" "--config" "Path to config.ini or conf.d directory (required)"]
-   ["-D" "--debug" "Enable debug mode" :default false :flag true]])
+(defn stop-puppetdb
+  "Callback to execute any necessary cleanup during a normal shutdown."
+  [context]
+  (log/info "Shutdown request received; puppetdb exiting.")
+  (shutdown-mq-broker context)
+  context)
 
-(def required-cli-options
-  [:config])
+(defn error-shutdown!
+  "Last-resort shutdown/cleanup code to execute when a fatal error has occurred."
+  [context]
+  (log/error "A fatal error occurred; shutting down all subsystems.")
+  (when-let [command-procs (context :command-procs)]
+    (log/info "Shutting down command processing threads.")
+    (doseq [cp command-procs]
+           (future-cancel cp)))
 
-(defn -main
-  [& args]
-  (let [[options _]                                (cli! args
-                                                      supported-cli-options
-                                                      required-cli-options)
-        initial-config                             {:debug (:debug options)}
-        {:keys [jetty database read-database global command-processing]
-            :as config}                            (conf/parse-config (:config options) initial-config)
+  (when-let [updater (context :updater)]
+    (log/info "Shutting down updater thread.")
+    (future-cancel updater))
+
+  (shutdown-mq-broker context))
+
+(defn add-max-framesize
+  "Add a maxFrameSize to broker url for activemq."
+  [config url]
+  (format "%s&wireFormat.maxFrameSize=%s&marshal=true" url (:max-frame-size config)))
+
+(defn start-puppetdb
+  [context config service-id add-ring-handler shutdown-on-error]
+  {:pre [(map? context)
+         (map? config)
+         (ifn? add-ring-handler)
+         (ifn? shutdown-on-error)]
+   :post [(map? %)
+          (every? (partial contains? %) [:broker :command-procs :updater])]}
+  (let [{:keys [jetty database read-database global command-processing]
+         :as config}                            (conf/process-config! config)
         product-name                               (:product-name global)
         update-server                              (:update-server global)
-        ;; TODO: revisit the choice of 20000 as a default value for event queries
-        event-query-limit                          (:event-query-limit global)
+        url-prefix                                 (:url-prefix global)
         write-db                                   (pl-jdbc/pooled-datasource database)
         read-db                                    (pl-jdbc/pooled-datasource (assoc read-database :read-only? true))
         gc-interval                                (get database :gc-interval)
@@ -250,68 +272,73 @@
         discard-dir                                (file mq-dir "discarded")
         globals                                    {:scf-read-db          read-db
                                                     :scf-write-db         write-db
-                                                    :command-mq           {:connection-string mq-addr
+                                                    :command-mq           {:connection-string (add-max-framesize command-processing mq-addr)
                                                                            :endpoint          mq-endpoint}
-                                                    :event-query-limit    event-query-limit
                                                     :update-server        update-server
-                                                    :product-name         product-name}]
-
-
-    (utils/log-deprecated-jdk)
+                                                    :product-name         product-name
+                                                    :url-prefix           url-prefix}]
 
     (when (version)
       (log/info (format "PuppetDB version %s" (version))))
 
-    ;; Add a shutdown hook where we can handle any required cleanup
-    (kitchensink/add-shutdown-hook! on-shutdown)
-
-    ;; Ensure the database is migrated to the latest version, and warn if it's
-    ;; deprecated. We do this in a single connection because HSQLDB seems to
-    ;; get confused if the database doesn't exist but we open and close a
+    ;; Ensure the database is migrated to the latest version, and warn
+    ;; if it's deprecated, log and exit if it's unsupported. We do
+    ;; this in a single connection because HSQLDB seems to get
+    ;; confused if the database doesn't exist but we open and close a
     ;; connection without creating anything.
     (sql/with-connection write-db
-      (scf-store/warn-on-db-deprecation!)
-      (migrate!))
+                         (scf-store/validate-database-version #(System/exit 1))
+                         (migrate!))
 
     ;; Initialize database-dependent metrics
     (pop/initialize-metrics write-db)
 
-    (let [error         (promise)
-          broker        (try
-                          (log/info "Starting broker")
-                          (mq/build-and-start-broker! "localhost" mq-dir command-processing)
-                          (catch java.io.EOFException e
-                            (log/error
-                              "EOF Exception caught during broker start, this "
-                              "might be due to KahaDB corruption. Consult the "
-                              "PuppetDB troubleshooting guide.")
-                            (throw e)))
+    (let [broker (try
+                   (log/info "Starting broker")
+                   (mq/build-and-start-broker! "localhost" mq-dir command-processing)
+                   (catch java.io.EOFException e
+                     (log/error
+                       "EOF Exception caught during broker start, this "
+                       "might be due to KahaDB corruption. Consult the "
+                       "PuppetDB troubleshooting guide.")
+                     (throw e)))
+          context (assoc context :broker broker)
           command-procs (let [nthreads (command-processing :threads)]
                           (log/info (format "Starting %d command processor threads" nthreads))
                           (vec (for [n (range nthreads)]
-                                 (future (with-error-delivery error
-                                           (load-from-mq mq-addr mq-endpoint discard-dir {:db write-db
-                                                                                          :catalog-hash-debug-dir (:catalog-hash-debug-dir global)}))))))
-          updater       (future (maybe-check-for-updates product-name update-server read-db))
-          web-app       (let [authorized? (if-let [wl (jetty :certificate-whitelist)]
-                                            (build-whitelist-authorizer wl)
-                                            (constantly true))
-                              app         (server/build-app :globals globals :authorized? authorized?)]
-                          (log/info "Starting query server")
-                          (future (with-error-delivery error
-                                    (jetty/run-jetty app jetty))))
-          job-pool      (mk-pool)]
+                                 (future (shutdown-on-error
+                                           service-id
+                                           #(load-from-mq
+                                             mq-addr
+                                             mq-endpoint
+                                             discard-dir
+                                             {:db                     write-db
+                                              :catalog-hash-debug-dir (:catalog-hash-debug-dir global)})
+                                           error-shutdown!)))))
+          context (assoc context :command-procs command-procs)
+          updater (future (maybe-check-for-updates product-name update-server read-db))
+          context (assoc context :updater updater)
+          _       (let [authorized? (if-let [wl (jetty :certificate-whitelist)]
+                                      (build-whitelist-authorizer wl)
+                                      (constantly true))
+                        app (server/build-app :globals globals :authorized? authorized?)]
+                    (log/info "Starting query server")
+                    (add-ring-handler (compojure/context url-prefix [] app) url-prefix))
+          job-pool (mk-pool)]
 
       ;; Pretty much this helper just knows our job-pool and gc-interval
       (let [gc-interval-millis (to-millis gc-interval)
             gc-task #(interspaced gc-interval-millis % job-pool)
-            db-maintenance-tasks [garbage-collect!
-                                  (when (pos? (to-secs node-ttl))
+            db-maintenance-tasks [(when (pos? (to-secs node-ttl))
                                     (partial auto-deactivate-nodes! node-ttl))
                                   (when (pos? (to-secs node-purge-ttl))
                                     (partial purge-nodes! node-purge-ttl))
                                   (when (pos? (to-secs report-ttl))
-                                    (partial sweep-reports! report-ttl))]]
+                                    (partial sweep-reports! report-ttl))
+                                  ;; Order is important here to ensure
+                                  ;; anything referencing an env or resource
+                                  ;; param is purged first
+                                  garbage-collect!]]
 
         (gc-task #(apply perform-db-maintenance! write-db (remove nil? db-maintenance-tasks)))
         (gc-task #(compress-dlo! dlo-compression-threshold discard-dir)))
@@ -321,15 +348,22 @@
         (when (kitchensink/true-str? enabled)
           (log/warn (format "Starting %s server on port %d" type port))
           (start-repl type host port)))
+      context)))
 
-      (let [exception (deref error)]
-        (doseq [cp command-procs]
-          (future-cancel cp))
-        (future-cancel updater)
-        (future-cancel web-app)
+(defservice puppetdb-service
+  "Defines a trapperkeeper service for PuppetDB; this service is responsible
+  for initializing all of the PuppetDB subsystems and registering shutdown hooks
+  that trapperkeeper will call on exit."
+  [[:ConfigService get-config]
+   [:WebserverService add-ring-handler]
+   [:ShutdownService shutdown-on-error]]
 
-        ;; Stop the mq the old-fashioned way
-        (mq/stop-broker! broker)
+  (start [this context]
+         (start-puppetdb context (get-config) (service-id this) add-ring-handler shutdown-on-error))
 
-        ;; Now throw the exception so the top-level handler will see it
-        (throw exception)))))
+  (stop [this context]
+        (stop-puppetdb context)))
+
+(defn -main
+  [& args]
+  (apply main args))
